# -*- coding:utf-8 -*-
import io

import numpy as np
import pandas as pd
import seaborn as sn
from matplotlib import pyplot as plt
from sklearn.metrics import classification_report, confusion_matrix, cohen_kappa_score

from TorchMiner import BasePlugin


class MultiClassesClassification(BasePlugin):
    """MultiClassesClassificationMetric
    This can be used directly if your loss function is torch.nn.CrossEntropy
    """
    requirements = ["TensorboardDrawer"]

    def __init__(
            self,
            accuracy=True,
            confusion=True,
            kappa_score=True,
            report=True,
            backend="TensorboardDrawer",
            forward=None
    ):
        super().__init__()
        self.accuracy = accuracy
        self.confusion_matrix = confusion
        self.kappa_score = kappa_score
        self.classification_report = report
        self.backend = backend  # TODO:Can be used for sheet recorder
        self.forward = forward

    # def before_init(self):
    #     self.create_sheet_column("latest_confusion_matrix", "Latest Confusion Matrix")
    #     self.create_sheet_column("kappa_score", "Kappa Score")
    #     self.create_sheet_column("accuracy", "Accuracy")

    def after_init(self, *args, **kwargs):
        self.recorder = self.miner.plugins.get(self.backend)

    def before_val_epoch_start(self, epoch, **ignore):
        # TODO:Recommended to attach these attributes to Miner Obj, for combination with Other Plugins
        self.predicts = []
        self.label = []

    def after_val_iteration_ended(self, predicts, data, **ignore):
        raw_output = predicts.detach().cpu().numpy()
        if self.forward:
            predicts, label = self.forward(raw_output, data)
        else:
            predicts = np.argmax(raw_output, axis=1)  # Batch first
            label = data[1].cpu().numpy()  # label
        self.predicts.append(predicts)
        self.label.append(label)

    def after_val_epoch_end(self, val_loss, **ignore):
        predicts = np.concatenate(self.predicts)
        label = np.concatenate(self.label)
        if self.accuracy:
            accuracy = (predicts == label).sum() / len(predicts)
            self.logger.info(f"Val Accuracy:{accuracy}")
            self.recorder.scalar("Val/Accuracy", accuracy)

        if self.confusion_matrix:
            matrix = confusion_matrix(label, predicts)
            df_cm = pd.DataFrame(matrix)
            svm = sn.heatmap(df_cm, annot=True, cmap="OrRd", fmt=".3g")
            figure = svm.get_figure()
            if val_loss < self.miner.lowest_val_loss:
                # Solution to store Matplotlib Figure in TensorBoardX
                # https://stackoverflow.com/questions/7821518/matplotlib-save-plot-to-numpy-array
                with io.BytesIO() as buff:
                    figure.savefig(buff, format='raw')
                    buff.seek(0)
                    data = np.frombuffer(buff.getvalue(), dtype=np.uint8)
                w, h = figure.canvas.get_width_height()
                im = data.reshape((int(h), int(w), -1))
                im = im.transpose((2, 0, 1))  # Change to CHW
                self.recorder.figure("Val/ConfusionMatrix", im)
            plt.close(figure)

        if self.kappa_score:
            kappa = cohen_kappa_score(label, predicts, weights="quadratic")
            self.recorder.scalar("Val/KappaScore", kappa)

        if self.classification_report:
            # TODO:Design a better way to output or reord classification report
            self.logger.info(classification_report(label, predicts))



# -*- coding:utf-8 -*-
from tensorboardX import SummaryWriter

from TorchMiner.plugins import BasePlugin


class TensorboardDrawer(BasePlugin):
    """To visualize everything in training process using tensorboard"""

    def __init__(self, input_to_model):
        super(TensorboardDrawer, self).__init__()
        self.input_to_model = input_to_model

    def prepare(self, miner, *args, **kwargs):
        super(TensorboardDrawer, self).prepare(miner)
        self.writer = SummaryWriter(log_dir=self.miner.experiment_dir)

    def after_init(self, **ignore):
        self.writer.add_graph(self.miner.model, self.input_to_model)

    def after_train_epoch_end(self, train_loss, epoch, **ignore):
        self.writer.add_scalar("Loss/train", train_loss, global_step=epoch)

    def after_val_epoch_end(self, val_loss, epoch, **ignore):
        self.writer.add_scalar("Loss/val", val_loss, global_step=epoch)

    # --- Outer APIs ---
    def scalar(self, label, value, epoch=None, **ignore):
        if not epoch:
            epoch = self.miner.current_epoch
        self.writer.add_scalar(label, value, epoch)

    def figure(self, label, value, epoch=None, **ignore):
        if not epoch:
            epoch = self.miner.current_epoch
        self.writer.add_image(label, value, epoch)
# def scalars(self, x, value, graph):
#     """
#     Add a scalar on a graph
#     :param x:
#     :param value:
#     :param graph:
#     :return:
#     """
#     if graph not in self.state:
#         self.state[graph] = 0
#     key = "{}/{}".format(self.miner.experiment, graph)
#     if isinstance(value, dict):
#         self.writer.add_scalars(key, value, self.state[graph])
#     else:
#         self.writer.add_scalar(key, value, self.state[graph])
#     self.state[graph] += 1



def seed_everything(seed):
    """
    Fix the seed for generating random numbers.
    :param seed:
    :return:
    """
    import torch
    from numpy import random

    torch.manual_seed(seed)
    # may lead to bad performance
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    random.seed(seed)

    if seed == 0:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def create_dir(directory, *args):
    """Create directory"""
    import os

    current_dir = directory
    for dir_name in args:
        # TODO:os.path.join
        current_dir = os.path.join(current_dir, dir_name)
    if not os.path.isdir(current_dir):
        os.mkdir(current_dir)



class BasePlugin:
    requirements = []

    def __init__(self):
        self.name = self.__class__.__name__
        self.miner = None
        self.logger = None  # Plugin Logger will be inited in prepare stage

    def prepare(self, miner, *args, **kwargs):
        # Can be used for Monkey Patch Operations
        self.miner = miner

    # Plugin Data Begin
    def load_state_dict(self, state):
        pass

    def state_dict(self):
        return {}

    # Plugin Data End

    def set_miner(self, miner):
        """
        This function is deprecated.
        :param miner:
        :return:
        """
        self.miner = miner

    # Hook Functions Begin
    def before_init(self, *args, **kwargs):
        self.logger = self.miner.get_logger(self.name)

    def after_init(self, *args, **kwargs):
        pass

    def before_train_epoch_start(self, *args, **kwargs):
        pass

    def before_train_iteration_start(self, *args, **kwargs):
        pass

    def after_train_iteration_end(self, *args, **kwargs):
        pass

    def after_train_epoch_end(self, *args, **kwargs):
        pass

    def before_val_epoch_start(self, *args, **kwargs):
        pass

    def before_val_iteration_start(self, *args, **kwargs):
        pass

    def after_val_iteration_ended(self, *args, **kwargs):
        pass

    def after_val_epoch_end(self, *args, **kwargs):
        pass

    def before_checkpoint_persisted(self, *args, **kwargs):
        pass

    def after_checkpoint_persisted(self, *args, **kwargs):
        pass

    def before_quit(self, *args, **kwargs):
        pass
    # Hook Functions end

    # def print_txt(self, printable, name):
    #     with open(self.plugin_file(f"{name}.txt"), "a") as f:
    #         print(
    #             f"================ Epoch {self.current_epoch} ================\n",
    #             file=f,
    #         )
    #         print(printable, file=f)
    #         print("\n\n", file=f)

    # @property
    # def plugin_dir(self):
    #     if hasattr(self, "_plugin_dir"):
    #         return getattr(self, "_plugin_dir")
    #
    #     plugin_dir = os.path.join(self.code_dir, self.__class__.__name__)
    #     try:
    #         os.mkdir(plugin_dir)
    #     except FileExistsError:
    #         pass
    #     self._plugin_dir = plugin_dir
    #     return self._plugin_dir
    #
    # def plugin_file(self, name):
    #     return os.path.join(self.plugin_dir, name)

    # def create_sheet_column(self, key, name):
    #     self.miner.create_sheet_column(f"{self.prefix}{key}", f"{self.prefix}{name}")
    #
    # def update_sheet(self, key, value):
    #     self.miner.update_sheet(f"{self.prefix}{key}", value)
    #
    # def scalars(self, values, graph):
    #     return self.miner.drawer.scalars(
    #         self.miner.current_epoch, values, f"{self.prefix}{graph}"
    #     )


class PluginManager:
    def __init__(self, miner, plugins: list):
        self.miner = miner
        if plugins:
            self.plugins = plugins
            self.plugin_names = [i.name for i in self.plugins]
        else:
            self.plugins = []
        self.maper = dict(zip(self.plugin_names, self.plugins))
        self.prepare()  # Prepare is the earliest
        self.logger = miner.get_logger("PluginManager")
        self.check_requirements()

    def check_requirements(self):
        """
        Check Requirements of each Plugins
         - Requirements are set in restrict string
        :return:
        """
        error = False
        for p in self.plugins:
            self.logger.debug(f"Checking Requirements of {p.name}.")
            for r in p.requirements:
                if r not in self.plugin_names:
                    self.logger.error(f"Requirement {r} of {p.name} is not Find.")
                    error = True
        if not error:
            self.logger.info("Successfully Passed Plugin Requirements Check with no Errors.")
        # TODO: Try to import Unmet needs

    def status(self):
        """
        Print the Status of all registered Plugins
        :return:
        """
        self.logger.info(f"Registered Plugins:{self.plugins}")

    def register(self):
        """
        Register Plugins from a given module list
        :return:
        """

    def call(self, hook, **payload):
        """
        Call Hook Functions
        :return:
        """
        for plugin in self.plugins:
            getattr(plugin, hook)(miner=self, **payload)  # !!! `miner=self` is totally different with just `self`

    def prepare(self):
        """
        prepare a given Plugin
        :return:
        """
        for p in self.plugins:
            p.prepare(self.miner)

    def load(self, checkpoint):
        # load plugin states
        for plugin in self.plugins:
            key = f"__plugin.{plugin.__class__.__name__}__"
            plugin.load_state_dict(checkpoint.get(key, {}))

    def save(self):
        temp = {}
        for p in self.plugins:
            temp[f"__plugin.{p.__class__.__name__}__"] = p.state_dict()
        return temp

    def get(self, name):
        """
        Get A Plugin From Plugin Manager.
        :param string name: Plugin name
        :return:
        """
        return self.maper[name]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['torchminer',
 'torchminer.plugins',
 'torchminer.plugins.Drawer',
 'torchminer.plugins.Logger',
 'torchminer.plugins.Metrics',
 'torchminer.plugins.Sheet']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.4.0,<9.0.0',
 'google-api-python-client>=2.31.0,<3.0.0',
 'ipython>=7.18.0,<8.0.0',
 'matplotlib>=3.5.0,<4.0.0',
 'pandas>=1.3.4,<2.0.0',
 'pylint>=2.12.1,<3.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'sklearn>=0.0,<0.1',
 'tensorboardX>=2.4.1,<3.0.0',
 'torch>=1.8.0,<2.0.0',
 'tqdm>=4.50.0,<5.0.0']

setup_kwargs = {
    'name': 'torchminer',
    'version': '0.4.1',
    'description': 'Run Torch With A Simple Miner',
    'long_description': "This Project is Forked From [MineTorch](https://github.com/louis-she/minetorch).\n\nPublished on [pypi](https://pypi.org/project/torchminer/)\n\nPackaged Using [Poetry](https://python-poetry.org/)\n\n# Description\nTorchMiner is designed to automatic process the training ,evaluating and testing process for PyTorch DeepLearning,with a simple API.\n\nYou can access all Functions of MineTorch simply use `Miner`.\n\n## Project ToDo\n [!] compatible with paddlepaddle\n\n [!!!] Test Cases\n \n [!] Add A thread to accept CLI input when training\n \n [!] Abstract Miner process, for easier patches\n \n [] Abstract Plugin Manager\n\n [] Move ***Drawer*** Operations Outside of Miner as a Plugin\n \n [] A Plugin that can record every output of network for future analysis\n \n [] Add Plugin Able And Disable Stat\n \n [] Move Miner Options to yaml File, Add Config Class\n \n Now Plugins only supports output functions, they can't modify or change the data of the Miner class.Any Ideas? I am glad to know.\n \n [] Write about my design concept\n \n Critical \n \n [] Deal About the input size problem, such as Batch-first...",
    'author': 'InEase',
    'author_email': 'inease28@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



# -*- coding:utf-8 -*-
import _pickle as pickle
import os

from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure

from TorchMiner.plugins import BasePlugin


class MatplotlibDrawer(BasePlugin):
    colors = [
        "blue",
        "orange",
        "green",
        "red",
        "purple",
        "brown",
        "pink",
        "gray",
        "olive",
        "cyan",
    ]

    def prepare(self, miner, *args, **kwargs):
        super(MatplotlibDrawer, self).prepare(miner)
        self.graph_dir = os.path.join(
            self.miner.alchemistic_directory, self.miner.experiment, "graphs"
        )
        self.data_file = os.path.join(self.graph_dir, ".graphs.pickle")
        if not os.path.isdir(self.graph_dir):
            os.mkdir(self.graph_dir)

        # Load from previous
        # TODO: Need to add into state_dict?
        if os.path.isfile(self.data_file):
            with open(self.data_file, "rb") as f:
                self.graph_data = pickle.load(f)

    def _update_state(self, x, values, graph):
        if graph not in self.state or not isinstance(self.state[graph], dict):
            self.state[graph] = {}
        for key in values:
            if key not in self.state[graph]:
                self.state[graph][key] = {}
            self.state[graph][key][x] = values[key]
        with open(self.data_file, "wb") as f:
            pickle.dump(self.state, f)

    def _save_png(self, graph):
        png_file = os.path.join(self.graph_dir, graph + ".png")
        fig = Figure()
        FigureCanvas(fig)
        ax = fig.add_subplot(1, 1, 1)
        ax.grid(True)
        for index, curve in enumerate(self.state[graph]):
            ax.plot(
                *zip(*sorted(self.state[graph][curve].items())),
                label=curve,
                color=self.colors[index]
            )

        ax.legend(loc="upper left")
        fig.savefig(png_file, facecolor="#F0FFFC")
        return png_file

    def scalars(self, x, values, graph):
        """Add a scalar on a graph

        Args:
            value (dict):
                scalars to put on the graph
            graph (string):
                graph name
        """
        self._update_state(x, values, graph)
        return self._save_png(graph)
